/*
  AntiRight
  (c) 2003-2004 Jeffrey Bedard
  jefbed@e-list.net
   
  This file is part of AntiRight.
   
  AntiRight is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
   
  AntiRight is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
   
  You should have received a copy of the GNU General Public License
  along with AntiRight; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include <assert.h>
#include <stdio.h>
#include "arshell-x.h"

enum
  {
    FALSE, TRUE
  };

struct arshell_color
{
  int arshell_color_scale;
  double scale;
};

struct arshell_row
{
  int rows;
  int current_column;
  int current_row;
};

struct arshell_text
{
  char* value;
  int widget_id;
  int field_counter;
  char field_string[BUFSIZ];
};

struct arshell_flags
{
  int command_field;
  int work_area;
  int exit_on_action;
  int status_area;
  int menubar;
  int color_selector;
};

enum
  {
    MAX_WIDGETS=BUFSIZ,
    MAX_CHARS=256
  };

enum
  {
    BUTTON, LABEL, TEXT, TEXT_FIELD, MENU_BUTTON, UPDATER, SCALE, 
    COLOR_SELECTOR
  };

struct Widget
{
  int x, y;
  char command[MAX_CHARS];
  char label[MAX_CHARS];
  int type;
  int button_pressed_flag;
};

struct arshell_data
{
  struct arshell_flags flags;
  struct arshell_text text;
  struct arshell_color color;
  struct arshell_row row;
  struct Widget widgets[MAX_WIDGETS];
  int widgets_int;
  int widget_width;
  int char_pixel_width;
};

struct arshell_data arshell;

void
arshell_data_construct(struct arshell_data* arshell)
{
  arshell->row.rows=6;
  arshell->color.scale=3.282;
  arshell->widget_width=96;
  arshell->char_pixel_width=7;
}

struct Graphics
{
  Display* display;
  Window window;
  GC gc;  
};

struct Graphics GUI;

void
initialize_Graphics(int x, int y, int width, int height)
{
  GUI.display=XOpenDisplay((char*)getenv("DISPLAY"));
  assert(GUI.display != NULL);
  GUI.window=XCreateSimpleWindow(GUI.display, DefaultRootWindow(GUI.display),
				 x, y, 
				 width, height,
				 2,
				 BlackPixel(GUI.display, 
					    DefaultScreen(GUI.display)),
				 WhitePixel(GUI.display, 
					    DefaultScreen(GUI.display)));
  GUI.gc=XCreateGC(GUI.display, GUI.window, 0, NULL);
  XStoreName(GUI.display, GUI.window, "AntiRight");
}
void
show_window()
{
  XMapWindow(GUI.display, GUI.window);  
}

enum
  {
    FRAME_IN, FRAME_OUT
  };

void
arshell_show_menubar()
{
  create_widget(MENU_BUTTON, 0, 16, "ACE -A File_Menu");
  create_widget(MENU_BUTTON, arshell.widget_width*1, 16, "ACE -A View_Menu");
  create_widget(MENU_BUTTON, arshell.widget_width*2, 16, "ACE -A Help_Menu");
}

unsigned long
arshell_get_pixel(int red, int green, int blue)
{
  XColor color;
  Colormap colormap=DefaultColormap(GUI.display, DefaultScreen(GUI.display));
  color.flags=DoRed|DoGreen|DoBlue;
  color.red=red*256;
  color.green=green*256;
  color.blue=blue*256;
  XAllocColor(GUI.display, colormap, &color);
  return(color.pixel);
}

void
arshell_set_background(int red, int green, int blue)
{
  XSetWindowBackground(GUI.display, GUI.window, 
		       arshell_get_pixel(red, green, blue));
}
void
arshell_set_foreground(int red, int green, int blue)
{
  XSetForeground(GUI.display, GUI.gc,
		 arshell_get_pixel(red, green, blue));
}

void
draw_top_shadow(int border_width, int x, int y, int width, int height)
{
  XDrawLine(GUI.display, GUI.window, GUI.gc,
	    x+border_width, y-(height-border_width),
	    x+(width-border_width), y-(height-border_width));
}
void
draw_bottom_shadow(int border_width, int x, int y, int width, int height)
{
  XDrawLine(GUI.display, GUI.window, GUI.gc,
	    x+border_width, y-border_width,
	    x+(width-border_width), y-border_width);
}
void
draw_left_shadow(int border_width, int x, int y, int width, int height)
{
  XDrawLine(GUI.display, GUI.window, GUI.gc,
	    x+border_width, y-(height-border_width),
	    x+border_width, y-border_width);
}
void
draw_right_shadow(int border_width, int x, int y, int width, int height)
{
  XDrawLine(GUI.display, GUI.window, GUI.gc,
	    x+(width-border_width), y-(height-border_width),
	    x+(width-border_width), y-border_width);
}
void
draw_frame(int style, 
	   int border_width, 
	   int x, int y, 
	   int width, int height)
{
  if(style==FRAME_OUT)
    arshell_set_foreground(255, 255, 255);
  else if(style==FRAME_IN)
    arshell_set_foreground(0, 0, 0);
  draw_top_shadow(border_width, x, y, width, height);
  draw_left_shadow(border_width, x, y, width, height);
  if(style==FRAME_OUT)
    arshell_set_foreground(0, 0, 0);
  else if(style==FRAME_IN)
    arshell_set_foreground(255, 255, 255);
  draw_bottom_shadow(border_width, x, y, width, height);
  draw_right_shadow(border_width, x, y, width, height);
}

void
draw_scale(int counter, int border_width)
{
  int height=16;
  int width=arshell.widget_width;
  int x=arshell.widgets[counter].x;
  int y=arshell.widgets[counter].y;
  draw_frame(FRAME_IN, border_width, x, y, width, height);
  draw_frame(FRAME_OUT, border_width, 
	     x+border_width+atoi(arshell.widgets[counter].command), 
	     y-border_width, 
	     16, height-border_width*2);
}
void
create_color_selector()
{
  arshell.flags.color_selector=TRUE;
  create_rowed_widget(SCALE, "COLOR_SELECTOR_RED");
  create_rowed_widget(SCALE, "COLOR_SELECTOR_GREEN");
  create_rowed_widget(SCALE, "COLOR_SELECTOR_BLUE");
  create_rowed_widget(LABEL, "COLOR_SELECTOR_LABEL");
  create_rowed_widget(BUTTON, "COLOR_SELECTOR_SELECT");
}

void
draw_button(int counter, int border_width)
{
  int height=16;
  int width=arshell.widget_width;
  int x=arshell.widgets[counter].x;
  int y=arshell.widgets[counter].y;
  
  if(arshell.widgets[counter].button_pressed_flag==FALSE)
    arshell_set_foreground(255, 255, 255);
  else
    arshell_set_foreground(0, 0, 0);
  draw_top_shadow(border_width, x, y, width, height);
  if(arshell.widgets[counter].type!=MENU_BUTTON)
    draw_left_shadow(border_width, x, y, width, height);
  if(arshell.widgets[counter].button_pressed_flag==FALSE)
    arshell_set_foreground(0, 0, 0);
  else
    arshell_set_foreground(255, 255, 255);
  draw_bottom_shadow(border_width, x, y, width, height);
  if(arshell.widgets[counter].type!=MENU_BUTTON)
    draw_right_shadow(border_width, x, y, width, height);
}

void
draw_label(int counter, int border_width)
{
  arshell_set_foreground(0, 0, 0);
  XDrawString(GUI.display, GUI.window, GUI.gc, 
	      arshell.widgets[counter].x+4, 
	      arshell.widgets[counter].y-border_width-2,
	      arshell.widgets[counter].label, 
	      strlen(arshell.widgets[counter].label));
}

void
draw_updater(int counter, int border_width)
{
  char* label_string=antiright_pipe_read(arshell.widgets[counter].command);
  strncpy(arshell.widgets[counter].label, label_string, MAX_CHARS);
  free(label_string);
  draw_label(counter, border_width);
}

void
draw_text_field(int counter, int border_width)
{
  int width=arshell.widget_width;
  int height=16;
  
  arshell_set_foreground(255, 255, 255);
  XFillRectangle(GUI.display, GUI.window, GUI.gc,
		 arshell.widgets[counter].x, arshell.widgets[counter].y-16,
		 width, height);
  arshell_set_foreground(111, 109, 107);
  XDrawRectangle(GUI.display, GUI.window, GUI.gc,
		 arshell.widgets[counter].x, arshell.widgets[counter].y-16,
		 width, height);
  draw_bottom_shadow(border_width, 
		     arshell.widgets[counter].x, arshell.widgets[counter].y, 
		     width, height);
}

void
update_display()
{
  int counter;
  int border_width=1;
  XSetLineAttributes(GUI.display, GUI.gc, 
		     border_width, LineSolid, CapRound, JoinRound);
  XClearWindow(GUI.display, GUI.window);
  arshell_set_background(222, 219, 214);
  for(counter=arshell.widgets_int; counter > 0; counter--)
    {
      if(arshell.widgets[counter].type==BUTTON 
	 | arshell.widgets[counter].type==MENU_BUTTON)
	  draw_button(counter, border_width);
      if(arshell.widgets[counter].type==LABEL 
	 | arshell.widgets[counter].type==BUTTON 
	 | arshell.widgets[counter].type==MENU_BUTTON)
	{
	  if(!strcmp(arshell.widgets[counter].command, 
		     "COLOR_SELECTOR_SELECT"))
	    strncpy(arshell.widgets[counter].label, "Select", MAX_CHARS);
	  if(!strcmp(arshell.widgets[counter].command, "COLOR_SELECTOR_LABEL")
	     | !strcmp(arshell.widgets[counter].label, "Color"))
	    {
	      strncpy(arshell.widgets[counter].label, "Color", MAX_CHARS);
	      arshell_set_foreground(abs
				     (atoi(arshell.widgets[counter-3].command)
				      *arshell.color.scale),
				     abs
				     (atoi(arshell.widgets[counter-2].command)
				      *arshell.color.scale),
				     abs
				     (atoi(arshell.widgets[counter-1].command)
				      *arshell.color.scale));
	      XFillRectangle(GUI.display, GUI.window, GUI.gc,
			     arshell.widgets[counter].x, 
			     arshell.widgets[counter].y-16,
			     arshell.widget_width, 16);
	      arshell_set_foreground(0, 0, 0);
	    }
	  draw_label(counter, border_width);
	}
      else if(arshell.widgets[counter].type==TEXT_FIELD)
	draw_text_field(counter, border_width);
      else if(arshell.widgets[counter].type==UPDATER)
	draw_updater(counter, border_width);
      else if(arshell.widgets[counter].type==SCALE)
	draw_scale(counter, border_width);
    }
}

void
create_widget(int type, int x, int y, char* command)
{
  arshell.widgets_int++;
  if(arshell.widgets_int<MAX_WIDGETS)
    {
      arshell.widgets[arshell.widgets_int].type=type;
      arshell.widgets[arshell.widgets_int].x=x;
      arshell.widgets[arshell.widgets_int].y=y;
      strncpy(arshell.widgets[arshell.widgets_int].command, command, 
	      MAX_CHARS);
      if(type==BUTTON | type==MENU_BUTTON | type==LABEL)
	{
	  char* command_string;
	  char* label_string=
	    (char*)antiright_beautified_label
	    (arshell.widgets[arshell.widgets_int].command);
	  int label_length;

	  if(type==LABEL)
	    label_length=MAX_CHARS;
	  else
	    label_length=abs(arshell.widget_width/arshell.char_pixel_width);
	  
	  strncpy(arshell.widgets[arshell.widgets_int].label, label_string, 
		  label_length);

	  free(label_string);
	}
      else if(type==TEXT_FIELD)
	{
	  arshell.text.widget_id=arshell.widgets_int;
	}
    }
  else
    {
      perror("Too many widgets.  ");
    }
}

void
cleanup_Graphics()
{
  XFreeGC(GUI.display, GUI.gc);
  XCloseDisplay(GUI.display);
}

void
check_commands_at_button(int x, int y)
{
  int counter;
  for(counter=arshell.widgets_int; counter > 0; counter--)
    {
      if(((arshell.widgets[counter].x < x) 
	  && (arshell.widgets[counter].x+arshell.widget_width > x))
	 && ((arshell.widgets[counter].y-16 < y) 
	     && (arshell.widgets[counter].y > y)))
	{
	  if(arshell.widgets[counter].type==SCALE)
	    {
	      if(x-arshell.widgets[counter].x+1<80)
		sprintf(arshell.widgets[counter].command, 
			"%d", x-arshell.widgets[counter].x, MAX_CHARS);
	    }
	  else
	    {
	      arshell.widgets[counter].button_pressed_flag=TRUE;
	      if(!strcmp(arshell.widgets[counter].command, 
			 "COLOR_SELECTOR_SELECT"))
		{
		  printf("rgb:%x/%x/%x", 
			 abs(atoi(arshell.widgets[counter-4].command)
			     *arshell.color.scale),
			 abs(atoi(arshell.widgets[counter-3].command)
			     *arshell.color.scale),
			 abs(atoi(arshell.widgets[counter-2].command)
			     *arshell.color.scale));
		  fflush(stdout);			 
		}
	      else
		antiright_system(arshell.widgets[counter].command);
	      if(arshell.flags.exit_on_action)
		{
		  cleanup_Graphics();
		  exit(0);
		}
	    }
	  
	}
    }
}

void
unpush_buttons()
{
  int counter;
  for(counter=arshell.widgets_int; counter > 0; counter--)
    arshell.widgets[counter].button_pressed_flag=FALSE;
}


void
event_loop()
{
  int text_y_int=16, text_x_int=0;
  int run_event_loop=TRUE;
  XSelectInput(GUI.display, GUI.window, 
	       ExposureMask | KeyPressMask | StructureNotifyMask 
	       | ButtonPressMask | ButtonReleaseMask | Button1MotionMask);
  while(run_event_loop==TRUE)
    {
      XEvent event;
      int size_int=32;
      char keyname_char[size_int];
      KeySym key;
      XComposeStatus compose_status;

      keyname_char[0]=0;
      
      XNextEvent(GUI.display, &event);
      switch(event.type)
	{
	case Expose:
	  update_display();
	  break;
	case ConfigureNotify:
	  update_display();
	  break;
	case KeyPress:
	  XLookupString(&event.xkey, keyname_char, size_int, 
			&key, &compose_status);

	  if(key==XK_Escape)
	    run_event_loop=FALSE;
	  if(arshell.flags.work_area | arshell.flags.command_field)
	    {
	      arshell.text.field_string[arshell.text.field_counter]=
		keyname_char[0];
	      arshell.text.field_counter++;
	      arshell.text.field_string[arshell.text.field_counter]='\0';
	      
	      XDrawString(GUI.display, GUI.window, GUI.gc, 
			  arshell.widgets[arshell.text.widget_id].x, 
			  arshell.widgets[arshell.text.widget_id].y-2,
			  arshell.text.field_string, 
			  arshell.text.field_counter);
	      if(key==XK_Return)
		{
		  if(arshell.flags.command_field)
		    {
		      printf("%s %s",
			     arshell.widgets[arshell.text.widget_id-1].command,
			     arshell.text.field_string);
		      fflush(stdout);
		      run_event_loop=FALSE;
		    }
		}
	    }
	  break;
	case ButtonPress:
	  check_commands_at_button(event.xbutton.x, event.xbutton.y);
	  update_display();
	  break;
	case ButtonRelease:
	  unpush_buttons();
	  update_display();
	  break;
	case MotionNotify:
	  check_commands_at_button(event.xbutton.x, event.xbutton.y);
	  update_display();
	  break;
	}
    }
}

void
create_rowed_widget(int type, char* command)
{
  if(arshell.row.current_row==arshell.row.rows)
    {
      arshell.row.current_column++;
      arshell.row.current_row=1;
    }
  else
    {
      arshell.row.current_row++;
    }
  if(arshell.flags.menubar)
    {
      create_widget(type, 
		    arshell.row.current_column*arshell.widget_width, 
		    arshell.row.current_row*16+16, 
		    command);
    }
  else
    {
      create_widget(type, 
		    arshell.row.current_column*arshell.widget_width, 
		    arshell.row.current_row*16, 
		    command);
    }
}

void
create_entry_widget(char* label_string)
{
  create_rowed_widget(LABEL, label_string);
  create_rowed_widget(TEXT_FIELD, NULL);
}

void
set_geometry(int* arg_counter, char** argv)
{
  int char_counter_int;
  int reached_x_flag=FALSE;
  int reached_width_flag=FALSE, reached_height_flag=FALSE;
  char tmp_string[8];
  int current_char_int=0;
  int item_begin_char_int=0;
  XWindowChanges changes;
  for(char_counter_int=0; 
      char_counter_int < strlen(argv[(*arg_counter)]);
      char_counter_int++)
    {
      if(reached_width_flag==FALSE)
	{
	  tmp_string[item_begin_char_int]
	    =argv[(*arg_counter)][current_char_int];
	  item_begin_char_int++;
	  current_char_int++;
	  if(argv[(*arg_counter)][current_char_int]=='x')
	    {
	      reached_width_flag=TRUE;
	      tmp_string[item_begin_char_int]='\0';
	      item_begin_char_int=0;
	      changes.width=atoi(tmp_string);
	    }
	}
      else if(reached_height_flag==FALSE)
	{
	  current_char_int++;
	  tmp_string[item_begin_char_int]
	    =argv[(*arg_counter)][current_char_int];
	  item_begin_char_int++;
			    
	  if(argv[(*arg_counter)][current_char_int]=='+')
	    {
	      reached_height_flag=TRUE;
	      tmp_string[item_begin_char_int]='\0';
	      item_begin_char_int=0;
	      puts(tmp_string);
	      changes.height=atoi(tmp_string);
	    }
	}
      else if(reached_x_flag==FALSE)
	{
	  tmp_string[item_begin_char_int]
	    =argv[(*arg_counter)][current_char_int];
	  current_char_int++;
	  item_begin_char_int++;
	  if(argv[(*arg_counter)][current_char_int]=='+')
	    {
	      reached_x_flag=TRUE;
	      tmp_string[current_char_int]='\0';
	      item_begin_char_int=0;
	      changes.x=atoi(tmp_string);
	    }
	}
      else
	{
	  tmp_string[item_begin_char_int]
	    =argv[(*arg_counter)][current_char_int];
	  current_char_int++;
	  item_begin_char_int++;
	}
    }
  tmp_string[current_char_int]='\0';
  item_begin_char_int=0;
  changes.y=atoi(tmp_string);
  printf("x: %d y: %d width: %d: height: %d\n",
	 changes.x, changes.y, changes.width, changes.height);
  XConfigureWindow(GUI.display, GUI.window, 
		   CWX | CWY | CWWidth | CWHeight,
		   &changes);
  fflush(stdout);
}

void
arshell_resize_to_fit()
{
  XWindowChanges changes;
  changes.width=0;
  changes.height=arshell.row.rows*16;
  if(arshell.flags.work_area)
    {
      changes.height+=256;
      changes.width+=256;
    }
  if(arshell.flags.command_field)
    {
      changes.height+=16;
      changes.width+=256;
    }
  if(arshell.flags.status_area)
    changes.height+=16;
  changes.width
    +=arshell.row.current_column*arshell.widget_width+arshell.widget_width;
  XConfigureWindow(GUI.display, GUI.window, 
		   CWHeight | CWWidth,
		   &changes);
}


int
main(int argc, char** argv)
{
  arshell_data_construct(&arshell);
  initialize_Graphics(48, 48, 256, 256);
  if(argc>1)
    {
      int arg_counter;
      for(arg_counter=1; arg_counter < argc; arg_counter++)
	{
	  if(argv[arg_counter][0]=='-')
	    {
	      switch(argv[arg_counter][1])
		{
		case 'A':
		  switch(argv[arg_counter][2])
		    {
		    case 'C':
		      arshell.flags.exit_on_action=TRUE;
		      create_color_selector();
		      break;
		    case 'l':
		      arg_counter++;
		      create_rowed_widget(LABEL, argv[arg_counter]);
		      break;
		    case 'm':
		      switch(argv[arg_counter][3])
			{
			default:
			  arshell.flags.menubar=TRUE;
			  arshell_show_menubar();
			}
		    case 's':
		      arshell.flags.status_area=TRUE;
		      break;
		    case 't':
		      arg_counter++;
		      arshell.flags.command_field=TRUE;
		      create_rowed_widget(LABEL, argv[arg_counter]);
		      create_rowed_widget(TEXT_FIELD, "TEXT_FIELD");
		      /*create_entry_widget(argv[arg_counter]);*/
		      break;
		    case 'w':
		      arshell.flags.work_area=TRUE;
		      break;
		    }
		  break;		
		case 'E':
		  arshell.flags.exit_on_action=TRUE;
		  break;
		case 'g': /* This specifies the geometry.  */
		  arg_counter++;
		  set_geometry(&arg_counter, argv);
		  break;
		case 'r':
		  arg_counter++;
		  arshell.row.rows=atoi(argv[arg_counter]);
		  break;
		case 'U':
		  arg_counter++;
		  create_widget(UPDATER, 0, arshell.row.rows*16,
				argv[arg_counter]);
		case 'W':
		  arg_counter++;
		  XStoreName(GUI.display, GUI.window, argv[arg_counter]);
		case 'x':
		  break;
		default:
		  puts("usage:  arshell-x "
		       "[-Ep][-A[[C][l label][m[fevh] label][s][t label][w]]"
		       "[-r rows][-U command][-W title][button]");
		  exit(1);
		}
	    }
	  else
	    {
	      create_rowed_widget(BUTTON, argv[arg_counter]);
	    }
	}
    }

  arshell_resize_to_fit();
  show_window();
  event_loop();
  cleanup_Graphics();
  if(arshell.text.value!=NULL)
    {
      free(arshell.text.value);
    }
  return(0);
}
